/**
 * This program selects instances for the 2019 edition of the XCSP3 solvers competition.
 *
 * WARNING: this program is not generic enough. This means that it does not handle all possible cases. Besides, it is absolutely not optimal, nor well written.
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <vector>
#include <set>
#include <map>
#include <boost/algorithm/string.hpp>
#include <functional>
#include <random>
using namespace std;

constexpr int randomSeed=2019;

void error(const string &msg)
{
  cerr << msg << endl;
  exit(1);
}

// list of filenames prefix which form a specific series and that do not fit in the general case
set<string> seriesPrefix; // example: ={"XCSP17/Random"}; 

class Selection
{
public:
  struct Instance
  {
    string type;
    int id;
    float hardness;
    string filename;
    bool selected=false;

    enum SeriesType {SERIES_IS_DIR,SERIES_IS_PROBLEM};
    static SeriesType seriesType;
    
    string series() const
    {
      for(auto s: seriesPrefix)
        if(boost::starts_with(filename,s))
          return s;

      if(seriesType==SERIES_IS_PROBLEM)
        return filename.substr(0,filename.find("/",filename.find("/")+1)); // everything up to the second slash (the problem name)
      else
        if(seriesType==SERIES_IS_DIR)
          return filename.substr(0,filename.rfind("/")); // everything up to the last slash (the path)
        else
          throw runtime_error("seriesType is unknown");
    }
  };

  
  struct Category
  {
    string name;
    float prob;
    function<bool(const Instance &)> selector;

    int nbToSelect;
    vector<Instance *> availableInstances, selectedInstances;

    int nbSubs; // number of subcategories to consider
    function<int(const Instance &)> idSub; // return the index of the subcategory a given instance belongs to

    vector<vector<Instance *>> subs; // list of subcategories
    int nextSub=0;// next subcategory where we should pick an instance
    
    Category(const string &name, float prob, function<bool(const Instance &)> selector,
             int nbSubs, function<int(const Instance &)> idSub):
      name(name), prob(prob), selector(selector), nbSubs(nbSubs), idSub(idSub) {}

    void skipEmptySubs()
    {
      int round=0;

      if(nextSub>=nbSubs)
      {
        nextSub=0;
        ++round;
      }
      
      while(subs[nextSub].empty())
      {
        ++nextSub;
        if(nextSub>=nbSubs)
        {
          nextSub=0;
          if(++round>1)
            error("all sub-categories empty in category "+name);
        }
      }
    }
  };


  struct Series
  {
    string name;
    int selected=0;
    vector<Instance *> instances;
  };
  
protected:
  vector<Instance> instances;
  vector<Category> categories;
  map<string,Series> series;
  default_random_engine randomEngine{randomSeed};
  
public:
  /**
   * read the input file (psql output format). Ignores files which math a path in "excludePrefix"
   */
  void read(const string &filename, const initializer_list<string> &excludePrefix)
  {
    string line;
    ifstream in(filename);
    Instance inst;
    char sep;
    
    if(!in.good())
    {
      cerr << "could not find file " << filename << endl;
      exit(1);
    }
 
    getline(in,line);
    getline(in,line);
    while(in.good())
    {
      if(in.peek()=='(')
        break;
      
      in >> inst.type >> sep;
      if(in.fail())
        error("failed to read type");
      if(sep!='|')
        error("separator not found");

      in >> inst.id >> sep;
      if(in.fail())
        error("failed to read id");      
      if(sep!='|')
        error("separator not found");

      in >> inst.hardness >> sep;
      if(in.fail())
        error("failed to read hardness");
      if(sep!='|')
        error("separator not found");

      getline(in,inst.filename);
      if(in.fail())
        error("failed to read filename");
      boost::trim(inst.filename);

      bool ok=true;
      for(const string &prefix: excludePrefix)
        if(boost::starts_with(inst.filename,prefix))
        {
          ok=false;
          break;
        }

      if(ok)
        instances.push_back(inst);
    }   
  }

  // add a category
  void add(const string &name, float prob, function<bool(const Instance &)> selector,
           int nbSubs, function<int(const Instance &)> idSub)
  {
    categories.emplace_back(name,prob,selector,nbSubs,idSub);
  }

  // build the list of selected instances
  void select(const string &filename, const string &type, int nbToSelect, int maxPerSeries, int minPerNewSeries, const string &newSeriesPrefix)
  {
    assert(minPerNewSeries<=maxPerSeries);
    
    //reset
    series.clear();
    for(auto &cat: categories)
    {
      cat.availableInstances.clear();
      cat.selectedInstances.clear();
      cat.nbToSelect=static_cast<int>(cat.prob*nbToSelect+.5);
    }
      
    // classify instances
    for(auto &inst: instances)
    {
      if(inst.type!=type)
        continue;

      string seriesName=inst.series();
      series[seriesName].name=seriesName;
      series[seriesName].instances.push_back(&inst);

      for(auto &cat: categories)
        if(cat.selector(inst))
          cat.availableInstances.push_back(&inst);          
    }

    // print a summary of each categories
    cout << "filename=" << filename << ", type=" << type << endl;
    for(auto &cat: categories)
    {
      cout << "  " << cat.name << ": " << cat.availableInstances.size() << " instances available, "
	   << cat.nbToSelect << " instances to select.\n";

      map<string,int> perSeries;
      for(auto inst: cat.availableInstances)
        ++perSeries[inst->series()];

      int sum=0;
      for(const auto &[name,count]: perSeries)
      {
        cout << "    " << count << " instances in series " << name << endl;
        sum+=(count<maxPerSeries)?count:maxPerSeries;
      }
      cout << "  possible total: " << sum << endl;
    }

    // check if there are enough instances in each category
    for(int i=0;i<categories.size();++i)
    {
      auto &cat=categories[i];
      
      if(cat.nbToSelect>cat.availableInstances.size())
      {
	// not enough instances in the category
        // here, we implement an easy solution, but which is not general enough
	int missing=cat.nbToSelect-cat.availableInstances.size();
        int prev=i-1,next=i+1;

        cout << "  WARNING: category " << cat.name << " has not enough instances. Trying to select every instance...\n";
        
	cat.nbToSelect=cat.availableInstances.size();

        // we have no choice but to select every instance, if possible
        while(!cat.availableInstances.empty())
        {
          Instance *selected=cat.availableInstances.back();
          cat.availableInstances.pop_back();

          selected->selected=true;
          cat.selectedInstances.push_back(selected);
          checkSeriesLimit(selected,maxPerSeries);
        }

        if(cat.nbToSelect!=cat.selectedInstances.size())
        {
          cout << "    some series were disabled in category " << cat.name << ". Only " << cat.selectedInstances.size() << " instances selected instead of " << cat.nbToSelect << " expected." << endl;
          missing+=cat.nbToSelect-cat.selectedInstances.size();
        }
        
        cat.nbToSelect=0;
        
        while(prev>=0 && categories[prev].availableInstances.empty())
          --prev;
        while(next<categories.size() && categories[next].availableInstances.empty())
          ++next;

        cout << "  " << missing << " instances could not be selected in category " << cat.name << ". ";
          
        if(prev>=0 && next<categories.size())
        {
          cout <<  missing/2 << " additional instances will be picked in category "
               << categories[prev].name << " and " << missing-missing/2 << " in category "
               << categories[next].name << endl;
          categories[prev].nbToSelect+=missing/2;
          categories[next].nbToSelect+=missing-missing/2;
        }
        else
          if(prev>=0)
          {
            cout << "  " << missing << " additional instances will be picked in category "
                 << categories[prev].name <<endl;
            categories[prev].nbToSelect+=missing;
          }
          else
            if(next<categories.size())
            {
              cout << "  " << missing << " additional instances will be picked in category "
                   << categories[next].name <<endl;
              categories[next].nbToSelect+=missing;
            }
            else
              error("All categories empty! Cannot complete selection!");
      }
    }

    cout << "selection of fresh instances\n";

    // select fresh instances first. The selection is done in a very
    // simplified way. In 2019, we have only 2 new series and it's
    // just not worth the effort to do better.
    {
      bool hasFreshSeries=false;
      for(auto &[name,data]: series)
        if(boost::starts_with(name,newSeriesPrefix))
          {
            cout << "new series this year: " << name << endl;
            hasFreshSeries=true;
            
            vector<Instance *> list=data.instances;

            for(int n=0;n<minPerNewSeries;++n)
            {
               uniform_int_distribution<size_t> uni{0,list.size()-1};
               int rnd=uni(randomEngine);
               Instance *selected=list[rnd];
               selected->selected=true;
               list.erase(list.begin()+rnd);
               checkSeriesLimit(selected,maxPerSeries);
            }
          }
      
      // selected fresh instances must be removed from the lists
      for(auto &cat: categories)
      {
        for(auto it=cat.availableInstances.begin();it!=cat.availableInstances.end();)
          if((*it)->selected)
          {
            cat.selectedInstances.push_back(*it);
            cat.availableInstances.erase(it);
          }
          else
            ++it;
      }

      if(!hasFreshSeries)
        cout << "  no fresh instances found\n";
    }

    initSubCategories();

    cout << "general selection\n";
    
    // general selection
    int nbSelected=0;

    // we may have already some instances selected
    for(auto &cat: categories)
      nbSelected+=cat.selectedInstances.size();

    //cout << "nbSelected=" << nbSelected << endl;
    
    while(nbSelected<nbToSelect)
    {
      for(auto &cat: categories)
      {
        //cout << cat.name << ": selected=" << cat.selectedInstances.size() << ", nbToSelect=" << cat.nbToSelect << endl;
        if(cat.selectedInstances.size()>=cat.nbToSelect)
          continue; // enough!

        if(cat.availableInstances.empty())
          error("category "+cat.name+" became empty. Selection is aborted!");
          
        // pick an instance in the next sub
        cat.skipEmptySubs();
        auto &sub=cat.subs[cat.nextSub];

        assert(!sub.empty());
        
        uniform_int_distribution<size_t> uni{0,sub.size()-1};
        int rnd=uni(randomEngine);
        Instance *selected=sub[rnd];
        selected->selected=true;
        cat.selectedInstances.push_back(selected);
        sub.erase(sub.begin()+rnd);
        eraseAll(cat.availableInstances,selected);
          
        checkSeriesLimit(selected,maxPerSeries);
        
        ++cat.nextSub;
        cat.skipEmptySubs();
                      
        if(++nbSelected>=nbToSelect)
          break;
      }
    }


    // dump selection
    ofstream out(filename);
    bool sep=false;
    
    for(auto &cat: categories)
      for(auto inst: cat.selectedInstances)
      {
        if(sep)
          out << " ";
        out << inst->id;
        sep=true;
      }
    
    out.close();

    // to print meaningful statistics, restore availableInstances and subs
    for(auto &cat: categories)
      for(auto inst: cat.selectedInstances)
        cat.availableInstances.push_back(inst);
    
    initSubCategories();

    cout << "SUMMARY PER CATEGORY (" << filename << "):\n";
    for(auto &cat: categories)
    {
      cout << "  category " << cat.name << ": " << cat.selectedInstances.size() << " instances selected\n";
      vector<int> count(cat.nbSubs);
      for(auto inst: cat.selectedInstances)
        ++count[cat.idSub(*inst)];
      for(size_t i=0;i<count.size();++i)
        cout << "    sub-category " << i << ": " << count[i] << " instances selected out of " << cat.subs[i].size() << endl;
    }
    
    cout << "SUMMARY PER SERIES (" << filename << "):\n";
    for(auto &[name,data]: series)
      if(data.selected)
        cout << "    " << data.selected << " instances in series " << name << endl;
  }

protected:
  // erase each occurrence of elem in v
  void eraseAll(vector<Instance *> &v, Instance *elem)
  {
    auto dst=v.begin(),src=v.begin();
    while(src!=v.end())
      if(*src==elem)
        ++src;
      else
        *dst++=*src++;

    v.resize(dst-v.begin());
  }

  // erase each instance of v that belongs to the series
  void eraseSeries(vector<Instance *> &v, const string &series)
  {
    for(auto it=v.begin();it!=v.end();)
      if((*it)->series()==series)
        v.erase(it);
      else
        ++it;
  }
  
  void initSubCategories()
  {
    // create subcategories
    for(auto &cat: categories)
    {
      cat.subs.clear();
      cat.subs.resize(cat.nbSubs);
      for(auto inst: cat.availableInstances)
        cat.subs[cat.idSub(*inst)].push_back(inst);

      cat.nextSub=0;
      while(cat.subs[cat.nextSub].empty())
        ++cat.nextSub;
    }
  }

  // once instance "selected" is chosen, check if we must disable the remaining instances of the series because we've reached the maximum number of instances per series
  void checkSeriesLimit(Instance *selected, int maxPerSeries)
  {
    series[selected->series()].selected++;
    if(series[selected->series()].selected>=maxPerSeries)
    {
      // remove every instance in the same series
      string series=selected->series();
      cout << "  " << maxPerSeries << " instances selected in " << series << ": disabling series " << endl;
      for(auto &cat: categories)
      {
        eraseSeries(cat.availableInstances,series);
        for(auto &sub: cat.subs)
          eraseSeries(sub,series);
      }
    }
  }
};

Selection::Instance::SeriesType Selection::Instance::seriesType=Selection::Instance::SERIES_IS_DIR;

int main()
{
  Selection main,minicsp,minicop;

  Selection::Instance::seriesType=Selection::Instance::SERIES_IS_PROBLEM;
  // some instances in the XCSP18/ directory are in fact duplicates of instances in the XCSP17/ dir. These must be excluded. The list appears in the proceedings of the 2018 edition of the competition.
  main.read("hardness.txt",{
      "XCSP18/Bibd",
        "XCSP18/CarSequencing",
        "XCSP18/ColouredQueens",
        "XCSP18/Crossword",
        "XCSP18/Dubois",
        "XCSP18/Frb",
        "XCSP18/GolombRuler"
        "XCSP18/GracefulGraph",
        "XCSP18/GraphColoring",
        "XCSP18/Haystacks",
        "XCSP18/Knapsack",
        "XCSP18/Langford",
        "XCSP18/LowAutocorrelation",
        "XCSP18/MagicHexagon",
        "XCSP18/MagicSquare",
        "XCSP18/Mario",
        "XCSP18/PseudoBoolean-dec",
        "XCSP18/PseudoBoolean-opt",
        "XCSP18/QuadraticAssignment",
        "XCSP18/Quasigroups",
        "XCSP18/Quasigroups_mini",
        "XCSP18/Rcpsp",
        "XCSP18/SocialGolfers",
        "XCSP18/SportsScheduling",
        "XCSP18/StillLife",
        "XCSP18/StripPacking",
        "XCSP18/Subisomorphism",
        "XCSP18/TravellingSalesman"
        });
  
  main.add("easy",0.1,[](const Selection::Instance &inst) {return inst.hardness<=120;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-0)/(120-0+0.01));});
  main.add("medium",0.35,[](const Selection::Instance &inst) {return inst.hardness>120 && inst.hardness<=1800;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-120)/(1800-120+0.01));});
  main.add("hard",0.35,[](const Selection::Instance &inst) {return inst.hardness>1800 && inst.hardness<2400;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-1800)/(2400-1800+0.01));});
  main.add("open",0.2,[](const Selection::Instance &inst) {return inst.hardness>=2400;},1,[](const Selection::Instance &inst){return 0;});

  int nbInstances=300,maxPerSeries=15,minPerNewSeries=10;
  main.select("selection-main-csp.list","CSP",nbInstances,maxPerSeries,minPerNewSeries,"XCSP19/");
  cout << "\n-----------------------------------------------------\n\n";
  main.select("selection-main-cop.list","COP",nbInstances,maxPerSeries,minPerNewSeries,"XCSP19/");

  cout << "\n-----------------------------------------------------\n\n";
  
  Selection::Instance::seriesType=Selection::Instance::SERIES_IS_DIR;
  minicsp.read("hardness-minitrack.txt",{
      "XCSP18/Bibd",
        "XCSP18/CarSequencing",
        "XCSP18/ColouredQueens",
        "XCSP18/Crossword",
        "XCSP18/Dubois",
        "XCSP18/Frb",
        "XCSP18/GolombRuler"
        "XCSP18/GracefulGraph",
        "XCSP18/GraphColoring",
        "XCSP18/Haystacks",
        "XCSP18/Knapsack",
        "XCSP18/Langford",
        "XCSP18/LowAutocorrelation",
        "XCSP18/MagicHexagon",
        "XCSP18/MagicSquare",
        "XCSP18/Mario",
        "XCSP18/PseudoBoolean-dec",
        "XCSP18/PseudoBoolean-opt",
        "XCSP18/QuadraticAssignment",
        "XCSP18/Quasigroups",
        "XCSP18/Quasigroups_mini",
        "XCSP18/Rcpsp",
        "XCSP18/SocialGolfers",
        "XCSP18/SportsScheduling",
        "XCSP18/StillLife",
        "XCSP18/StripPacking",
        "XCSP18/Subisomorphism",
        "XCSP18/TravellingSalesman"
        });

  seriesPrefix={"XCSP17/Random","XCSP17/Crossword/Crossword-m1c-lex","XCSP17/Crossword/Crossword-m1c-ogd","XCSP17/Crossword/Crossword-m1c-uk","XCSP17/Crossword/Crossword-m1c-words","XCSP17/Subisomorphism/Subisomorphism-m1-si"};
  
  minicsp.add("easy",0.1,[](const Selection::Instance &inst) {return inst.hardness<=120;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-0)/(120-0+0.01));});
  minicsp.add("medium",0.35,[](const Selection::Instance &inst) {return inst.hardness>120 && inst.hardness<=1800;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-120)/(1800-120+0.01));});
  minicsp.add("hard",0.35,[](const Selection::Instance &inst) {return inst.hardness>1800 && inst.hardness<2400;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-1800)/(2400-1800+0.01));});
  minicsp.add("open",0.2,[](const Selection::Instance &inst) {return inst.hardness>=2400;},1,[](const Selection::Instance &inst){return 0;});

  nbInstances=200;
  maxPerSeries=10;
  minPerNewSeries=10;
  
  minicsp.select("selection-mini-csp.list","CSP",nbInstances,maxPerSeries,minPerNewSeries,"XCSP19/");


  cout << "\n-----------------------------------------------------\n\n";

  Selection::Instance::seriesType=Selection::Instance::SERIES_IS_DIR;
  minicop.read("hardness-minitrack.txt",{
      "XCSP18/Bibd",
        "XCSP18/CarSequencing",
        "XCSP18/ColouredQueens",
        "XCSP18/Crossword",
        "XCSP18/Dubois",
        "XCSP18/Frb",
        "XCSP18/GolombRuler"
        "XCSP18/GracefulGraph",
        "XCSP18/GraphColoring",
        "XCSP18/Haystacks",
        "XCSP18/Knapsack",
        "XCSP18/Langford",
        "XCSP18/LowAutocorrelation",
        "XCSP18/MagicHexagon",
        "XCSP18/MagicSquare",
        "XCSP18/Mario",
        "XCSP18/PseudoBoolean-dec",
        "XCSP18/PseudoBoolean-opt",
        "XCSP18/QuadraticAssignment",
        "XCSP18/Quasigroups",
        "XCSP18/Quasigroups_mini",
        "XCSP18/Rcpsp",
        "XCSP18/SocialGolfers",
        "XCSP18/SportsScheduling",
        "XCSP18/StillLife",
        "XCSP18/StripPacking",
        "XCSP18/Subisomorphism",
        "XCSP18/TravellingSalesman"
        });

  seriesPrefix={"XCSP17/Random","XCSP17/PseudoBoolean"};
  
  minicop.add("easy",0.3,[](const Selection::Instance &inst) {return inst.hardness<=120;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-0)/(120-0+0.01));});
  minicop.add("medium",0.3,[](const Selection::Instance &inst) {return inst.hardness>120 && inst.hardness<=1800;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-120)/(1800-120+0.01));});
  minicop.add("hard",0.1,[](const Selection::Instance &inst) {return inst.hardness>1800 && inst.hardness<2400;},10,[](const Selection::Instance &inst){return static_cast<int>(10*(inst.hardness-1800)/(2400-1800+0.01));});
  minicop.add("open",0.3,[](const Selection::Instance &inst) {return inst.hardness>=2400;},1,[](const Selection::Instance &inst){return 0;});

  nbInstances=180;
  maxPerSeries=20;
  minPerNewSeries=10;
  
  minicop.select("selection-mini-cop.list","COP",nbInstances,maxPerSeries,minPerNewSeries,"XCSP19/");
}
